/*=============================================================================
    Copyright (c) 2004 Reece H. Dunn

    Use, modification and distribution is subject to the Boost Software
    License, Version 1.0. (See accompanying file LICENSE_1_0.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
=============================================================================*/
#ifndef PHOENIX_STL_UTILITY_HPP
#define PHOENIX_STL_UTILITY_HPP

#include <boost/spirit/phoenix/function/function.hpp>

namespace boost { namespace phoenix
{
///////////////////////////////////////////////////////////////////////////////
//
//  std::pair accessor functions
//
//      Lazy functions for std::pair member access
//
//      These functions provide a mechanism for accessing the public members
//      of std::pair using lazy evaluation. For an overview of what is meant
//      by 'lazy evaluation', see the comments in operators.hpp and
//      functions.hpp.
//
//      To summarize, lazy functions are provided for data members:
//
//          first  second.
//
//      The lazy functions' names are the same as the corresponding data
//      member. Sample usage:
//
//      "Normal" version                 "Lazy" version
//      ----------------                 --------------
//      my_pair.first                    phoenix::first(arg1)
//      my_pair.second                   phoenix::second(arg1)
//
///////////////////////////////////////////////////////////////////////////////

namespace stl
{
///////////////////////////////////////////////////////////////////////////////
//
//  Metafunctions "first_type_of", "second_type_of".
//
//      These metafunctions define a typedef "type" that returns the type of
//      the first and second member in the pair construct respectively.
//
//      For example, "first_type_of<std::pair<char, int> >::type" is "char".
//
///////////////////////////////////////////////////////////////////////////////
    template <typename C>
    struct first_type_of
    {
        typedef typename C::first_type type;
    };

    template <typename C>
    struct second_type_of
    {
        typedef typename C::second_type type;
    };

///////////////////////////////////////////////////////////////////////////////
//
//  Lazy member function implementaions.
//
//      The structs below provide the guts of the implementation. Thereafter,
//      the corresponding lazy function itself is simply:
//
//          function<stl::first> const first = stl::first();
//
//      The structs provide a nested "apply" class template whose
//      "type" typedef enables the lazy function to ascertain the type
//      to be returned when it is invoked.
//
//      They also provide operator() member functions with signatures
//      corresponding to those of the underlying member function of
//      the STL container.
//
///////////////////////////////////////////////////////////////////////////////
    struct first
    {
        template <typename C>
        struct apply
        {
            typedef typename first_type_of<C>::type& type;
        };

        template <typename C>
        typename apply<C>::type
        operator()( C& c ) const
        {
            return c.first;
        }
     };

     struct second
     {
        template <typename C>
        struct apply
        {
            typedef typename second_type_of<C>::type& type;
        };

        template <typename C>
        typename apply<C>::type
        operator()( C& c ) const
        {
            return c.second;
        }
    };
} // namespace stl

///////////////////////////////////////////////////////////////////////////////
//
//  The lazy functions themselves.
//
///////////////////////////////////////////////////////////////////////////////
function<stl::first> const  first  = stl::first();
function<stl::second> const second = stl::second();

}} // namespace boost::phoenix

#endif // PHOENIX_STL_UTILITY_HPP
